# -*- coding: utf-8 -*-

from __future__ import division

import unittest
import numpy as np

from skyllh.physics.flux import (
    PowerLawFlux,
    CutoffPowerLawFlux,
    LogParabolaPowerLawFlux
)

from skyllh.core.py import float_cast


class TestPowerLawFlux(unittest.TestCase):
    def setUp(self):
        self.Phi0 = 1.5
        self.E0 = 2.5
        self.gamma = 2.7
        self.energy_unit = "GeV"
        self.power_law_flux = PowerLawFlux(self.Phi0, self.E0, self.gamma)

    def test_gamma(self):
        gamma = 2.6

        self.assertEqual(self.power_law_flux.gamma, self.gamma)

        self.power_law_flux.gamma = gamma

        self.assertEqual(self.power_law_flux.gamma, gamma)

    def test_math_function_str(self):
        self.assertEqual(self.power_law_flux.math_function_str, "dN/dE = {:.2e} * (E / {:.2e} {})^-{:.2f}".format(self.Phi0, self.E0, self.energy_unit, self.gamma))

    def test_call(self):
        E = 5
        flux = self.Phi0 * np.power(E / self.E0, -self.gamma)

        self.assertEqual(self.power_law_flux(E), flux)


class TestCutoffPowerLawFlux(unittest.TestCase):
    def setUp(self):
        self.Phi0 = 1.5
        self.E0 = 2.5
        self.gamma = 2.7
        self.Ecut = 2
        self.energy_unit = "GeV"
        self.cutoff_power_law_flux = CutoffPowerLawFlux(self.Phi0, self.E0, self.gamma, self.Ecut)

    def test_Ecut(self):
        Ecut = 2.5

        self.assertEqual(self.cutoff_power_law_flux.Ecut, self.Ecut)

        self.cutoff_power_law_flux.Ecut = Ecut

        self.assertEqual(self.cutoff_power_law_flux.Ecut, Ecut)

    def test_math_function_str(self):
        test_string = "dN/dE = {:.2e} * (E / {:.2e} {})^-{:.2f} * exp(-E / {:.2e} {})".format(self.Phi0, self.E0, self.energy_unit, self.gamma, self.Ecut, self.energy_unit)

        self.assertEqual(self.cutoff_power_law_flux.math_function_str, test_string)

    def test_call(self):
        E = 5
        flux = self.Phi0 * np.power(E / self.E0, -self.gamma) * np.exp(-E / self.Ecut)

        self.assertEqual(self.cutoff_power_law_flux(E), flux)


class TestLogParabolaPowerLawFlux(unittest.TestCase):
    def setUp(self):
        self.Phi0 = 1.5
        self.E0 = 2.5
        self.alpha = 1
        self.beta = 2
        self.energy_unit = "GeV"
        self.log_parabola_power_law_flux = LogParabolaPowerLawFlux(self.Phi0, self.E0, self.alpha, self.beta)

    def test_alpha(self):
        self.assertEqual(self.log_parabola_power_law_flux.alpha, self.alpha)

        alpha = 1.5
        self.log_parabola_power_law_flux.alpha = alpha

        self.assertEqual(self.log_parabola_power_law_flux.alpha, alpha)

    def test_beta(self):
        self.assertEqual(self.log_parabola_power_law_flux.beta, self.beta)

        beta = 2.5
        self.log_parabola_power_law_flux.beta = beta

        self.assertEqual(self.log_parabola_power_law_flux.beta, beta)

    def test_math_funciton_str(self):
        test_string = 'dN/dE = {:.2e} * (E / {:.2e} {})^(-({:.2e} + {:.2e} * log(E / {:.2e} {})))'.format(self.Phi0, self.E0, self.energy_unit, self.alpha, self.beta, self.E0, self.energy_unit)

        self.assertEqual(self.log_parabola_power_law_flux.math_function_str, test_string)

    def test_call(self):
        E = 5
        flux = self.Phi0 * np.power(E / self.E0, -self.alpha - self.beta * np.log(E / self.E0))

        self.assertEqual(self.log_parabola_power_law_flux(E), flux)


if(__name__ == '__main__'):
    unittest.main()



# -*- coding: utf-8 -*-

import abc
import itertools
import numpy as np

from skyllh.core.py import (
    issequenceof,
    float_cast,
    int_cast,
    get_smallest_numpy_int_type
)
from skyllh.core.dataset import Dataset, DatasetData
from skyllh.core.source_hypothesis import SourceHypoGroupManager
from skyllh.core.storage import DataFieldRecordArray
from skyllh.physics.flux import (
    get_conversion_factor_to_internal_flux_unit
)


class SignalGeneratorBase(object, metaclass=abc.ABCMeta):
    """This is the abstract base class for all signal generator classes in
    SkyLLH. It defines the interface for signal generators.
    """
    def __init__(self, src_hypo_group_manager, dataset_list, data_list,
                 *args, **kwargs):
        """Constructs a new signal generator instance.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The SourceHypoGroupManager instance defining the source hypothesis
            groups.
        dataset_list : list of Dataset instances
            The list of Dataset instances for which signal events should get
            generated for.
        data_list : list of DatasetData instances
            The list of DatasetData instances holding the actual data of each
            dataset. The order must match the order of ``dataset_list``.
        """
        super().__init__(*args, **kwargs)

        self.src_hypo_group_manager = src_hypo_group_manager
        self.dataset_list = dataset_list
        self.data_list = data_list

    @property
    def src_hypo_group_manager(self):
        """The SourceHypoGroupManager instance defining the source hypothesis
        groups.
        """
        return self._src_hypo_group_manager
    @src_hypo_group_manager.setter
    def src_hypo_group_manager(self, manager):
        if(not isinstance(manager, SourceHypoGroupManager)):
            raise TypeError(
                'The src_hypo_group_manager property must be an instance of '
                'SourceHypoGroupManager!')
        self._src_hypo_group_manager = manager

    @property
    def dataset_list(self):
        """The list of Dataset instances for which signal events should get
        generated for.
        """
        return self._dataset_list
    @dataset_list.setter
    def dataset_list(self, datasets):
        if(not issequenceof(datasets, Dataset)):
            raise TypeError(
                'The dataset_list property must be a sequence of Dataset '
                'instances!')
        self._dataset_list = list(datasets)

    @property
    def data_list(self):
        """The list of DatasetData instances holding the actual data of each
        dataset. The order must match the order of the ``dataset_list``
        property.
        """
        return self._data_list
    @data_list.setter
    def data_list(self, datas):
        if(not issequenceof(datas, DatasetData)):
            raise TypeError(
                'The data_list property must be a sequence of DatasetData '
                'instances!')
        self._data_list = datas

    def change_source_hypo_group_manager(self, src_hypo_group_manager):
        """Changes the source hypothesis group manager. Derived classes can
        reimplement this method but this method of the base class must still be
        called by the derived class.
        """
        self.src_hypo_group_manager = src_hypo_group_manager

    @abc.abstractmethod
    def generate_signal_events(self, rss, mean, poisson=True):
        """This abstract method must be implemented by the derived class to
        generate a given number of signal events.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService providing the random number
            generator state.
        mean : float
            The mean number of signal events. If the ``poisson`` argument is set
            to True, the actual number of generated signal events will be drawn
            from a Poisson distribution with this given mean value of signal
            events.
        poisson : bool
            If set to True, the actual number of generated signal events will
            be drawn from a Poisson distribution with the given mean value of
            signal events.
            If set to False, the argument ``mean`` specifies the actual number
            of generated signal events.

        Returns
        -------
        n_signal : int
            The number of generated signal events.
        signal_events_dict : dict of DataFieldRecordArray
            The dictionary holding the DataFieldRecordArray instancs with the
            generated signal events. Each key of this dictionary represents the
            dataset index for which the signal events have been generated.
        """
        pass


class SignalGenerator(SignalGeneratorBase):
    """This is the general signal generator class. It does not depend on the
    detector or source hypothesis, because these dependencies are factored out
    into the signal generation method. In fact the construction within this
    class depends on the construction of the signal generation method. In case
    of multiple sources the handling here is very suboptimal. Therefore the
    MultiSourceSignalGenerator should be used instead!
    """
    def __init__(self, src_hypo_group_manager, dataset_list, data_list,
                 *args, **kwargs):
        """Constructs a new signal generator instance.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The SourceHypoGroupManager instance defining the source groups with
            their spectra.
        dataset_list : list of Dataset instances
            The list of Dataset instances for which signal events should get
            generated for.
        data_list : list of DatasetData instances
            The list of DatasetData instances holding the actual data of each
            dataset. The order must match the order of ``dataset_list``.
        """
        super().__init__(
            *args,
            src_hypo_group_manager=src_hypo_group_manager,
            dataset_list=dataset_list,
            data_list=data_list,
            **kwargs)

        self._construct_signal_candidates()

    def _construct_signal_candidates(self):
        """Constructs an array holding pointer information of signal candidate
        events pointing into the real MC dataset(s).
        """
        n_datasets = len(self._dataset_list)
        n_sources = self._src_hypo_group_manager.n_sources
        shg_list = self._src_hypo_group_manager.src_hypo_group_list
        sig_candidates_dtype = [
            ('ds_idx', get_smallest_numpy_int_type((0, n_datasets))),
            ('ev_idx', get_smallest_numpy_int_type(
                [0]+[len(data.mc) for data in self._data_list])),
            ('shg_idx', get_smallest_numpy_int_type((0, n_sources))),
            ('shg_src_idx', get_smallest_numpy_int_type(
                [0]+[shg.n_sources for shg in shg_list])),
            ('weight', np.float64)
        ]
        self._sig_candidates = np.empty(
            (0,), dtype=sig_candidates_dtype, order='F')

        # Go through the source hypothesis groups to get the signal event
        # candidates.
        for ((shg_idx,shg), (j,(ds,data))) in itertools.product(
            enumerate(shg_list), enumerate(zip(self._dataset_list, self._data_list))):
            sig_gen_method = shg.sig_gen_method
            if(sig_gen_method is None):
                raise ValueError('No signal generation method has been '
                    'specified for the %dth source hypothesis group!'%(shg_idx))
            data_mc = data.mc
            (ev_indices_list, flux_list) = sig_gen_method.calc_source_signal_mc_event_flux(
                data_mc, shg
            )
            for (k, (ev_indices, flux)) in enumerate(zip(ev_indices_list, flux_list)):
                ev = data_mc[ev_indices]
                # The weight of the event specifies the number of signal events
                # this one event corresponds to for the given reference flux.
                # [weight] = GeV cm^2 sr * s * 1/(GeV cm^2 s sr)
                weight = ev['mcweight'] * data.livetime * 86400 * flux

                sig_candidates = np.empty(
                    (len(ev_indices),), dtype=sig_candidates_dtype, order='F'
                )
                sig_candidates['ds_idx'] = j
                sig_candidates['ev_idx'] = ev_indices
                sig_candidates['shg_idx'] = shg_idx
                sig_candidates['shg_src_idx'] = k
                sig_candidates['weight'] = weight

                self._sig_candidates = np.append(self._sig_candidates, sig_candidates)

        # Normalize the signal candidate weights.
        self._sig_candidates_weight_sum = np.sum(self._sig_candidates['weight'])
        self._sig_candidates['weight'] /= self._sig_candidates_weight_sum

    def change_source_hypo_group_manager(self, src_hypo_group_manager):
        """Recreates the signal candidates with the changed source hypothesis
        group manager.
        """
        super().change_source_hypo_group_manager(src_hypo_group_manager)

        self._construct_signal_candidates()

    def mu2flux(self, mu, per_source=False):
        """Translate the mean number of signal events `mu` into the
        corresponding flux. The unit of the returned flux is 1/(GeV cm^2 s).

        Parameters
        ----------
        mu : float
            The mean number of expected signal events for which to get the flux.
        per_source : bool
            Flag if the flux should be returned for each source individually
            (True), or as the sum of all these fluxes (False). The default is
            False.

        Returns
        -------
        mu_flux : float | (n_sources,)-shaped numpy ndarray
            The total flux for all sources (if `per_source = False`) that would
            correspond to the given mean number of detected signal events `mu`.
            If `per_source` is set to True, a numpy ndarray is returned that
            contains the flux for each individual source.
        """
        # Calculate the expected mean number of signal events for each source
        # of the source hypo group manager. For each source we can calculate the
        # flux that would correspond to the given mean number of signal events
        # `mu`. The total flux for all sources is then just the sum.

        # The ref_N variable describes how many total signal events are expected
        # on average for the reference fluxes.
        ref_N = self._sig_candidates_weight_sum

        # The mu_fluxes array is the flux of each source for mu mean detected
        # signal events.
        n_sources = self._src_hypo_group_manager.n_sources
        mu_fluxes = np.empty((n_sources,), dtype=np.float64)

        shg_list = self._src_hypo_group_manager.src_hypo_group_list
        mu_fluxes_idx_offset = 0
        for (shg_idx,shg) in enumerate(shg_list):
            fluxmodel = shg.fluxmodel
            # Calculate conversion factor from the flux model unit into the
            # internal flux unit GeV^-1 cm^-2 s^-1.
            toGeVcm2s = get_conversion_factor_to_internal_flux_unit(fluxmodel)
            for k in range(shg.n_sources):
                mask = ((self._sig_candidates['shg_idx'] == shg_idx) &
                        (self._sig_candidates['shg_src_idx'] == k))
                ref_N_k = np.sum(self._sig_candidates[mask]['weight']) * ref_N
                mu_flux_k = mu / ref_N * (ref_N_k / ref_N) * fluxmodel.Phi0*toGeVcm2s
                mu_fluxes[mu_fluxes_idx_offset + k] = mu_flux_k
            mu_fluxes_idx_offset += shg.n_sources

        if(per_source):
            return mu_fluxes

        mu_flux = np.sum(mu_fluxes)
        return mu_flux

    def generate_signal_events(self, rss, mean, poisson=True):
        """Generates a given number of signal events from the signal candidate
        monte-carlo events.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService providing the random number
            generator state.
        mean : float
            The mean number of signal events. If the ``poisson`` argument is set
            to True, the actual number of generated signal events will be drawn
            from a Poisson distribution with this given mean value of signal
            events.
        poisson : bool
            If set to True, the actual number of generated signal events will
            be drawn from a Poisson distribution with the given mean value of
            signal events.
            If set to False, the argument ``mean`` specifies the actual number
            of generated signal events.

        Returns
        -------
        n_signal : int
            The number of generated signal events.
        signal_events_dict : dict of DataFieldRecordArray
            The dictionary holding the DataFieldRecordArray instancs with the
            generated signal events. Each key of this dictionary represents the
            dataset index for which the signal events have been generated.
        """
        if(poisson):
            mean = rss.random.poisson(float_cast(
                mean, 'The mean argument must be castable to type of float!'))

        n_signal = int_cast(
            mean, 'The mean argument must be castable to type of int!')

        # Draw n_signal signal candidates according to their weight.
        sig_events_meta = rss.random.choice(
            self._sig_candidates,
            size=n_signal,
            p=self._sig_candidates['weight']
        )
        # Get the list of unique dataset and source hypothesis group indices of
        # the drawn signal events.
        # Note: This code does not assume the same format for each of the
        #       individual MC dataset numpy record arrays, thus might be a bit
        #       slower. If one could assume the same MC dataset format, one
        #       could gather all the MC events of all the datasets first and do
        #       the signal event post processing for all datasets at once.
        signal_events_dict = dict()
        ds_idxs = np.unique(sig_events_meta['ds_idx'])
        for ds_idx in ds_idxs:
            mc = self._data_list[ds_idx].mc
            ds_mask = sig_events_meta['ds_idx'] == ds_idx
            n_sig_events_ds = np.count_nonzero(ds_mask)

            data = dict(
                [(fname, np.empty(
                    (n_sig_events_ds,),
                    dtype=mc.get_field_dtype(fname))
                 ) for fname in mc.field_name_list])
            sig_events = DataFieldRecordArray(data, copy=False)

            fill_start_idx = 0
            # Get the list of unique source hypothesis group indices for the
            # current dataset.
            shg_idxs = np.unique(sig_events_meta[ds_mask]['shg_idx'])
            for shg_idx in shg_idxs:
                shg = self._src_hypo_group_manager.src_hypo_group_list[shg_idx]
                shg_mask = sig_events_meta['shg_idx'] == shg_idx
                # Get the MC events for the drawn signal events.
                ds_shg_mask = ds_mask & shg_mask
                shg_sig_events_meta = sig_events_meta[ds_shg_mask]
                n_shg_sig_events = len(shg_sig_events_meta)
                ev_idx = shg_sig_events_meta['ev_idx']
                # Get the signal MC events of the current dataset and source
                # hypothesis group.
                shg_sig_events = mc.get_selection(ev_idx)

                # Do the signal event post sampling processing.
                shg_sig_events = shg.sig_gen_method.signal_event_post_sampling_processing(
                    shg, shg_sig_events_meta, shg_sig_events)

                indices = np.indices((n_shg_sig_events,))[0] + fill_start_idx
                sig_events.set_selection(indices, shg_sig_events)

                #sig_events[fill_start_idx:fill_start_idx+n_shg_sig_events] = shg_sig_events
                fill_start_idx += n_shg_sig_events

            signal_events_dict[ds_idx] = sig_events

        return (n_signal, signal_events_dict)


class MultiSourceSignalGenerator(SignalGenerator):
    """More optimal signal generator for multiple sources.
    """
    def __init__(self, src_hypo_group_manager, dataset_list, data_list):
        """Constructs a new signal generator instance.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The SourceHypoGroupManager instance defining the source groups with
            their spectra.
        dataset_list : list of Dataset instances
            The list of Dataset instances for which signal events should get
            generated for.
        data_list : list of DatasetData instances
            The list of DatasetData instances holding the actual data of each
            dataset. The order must match the order of ``dataset_list``.
        """
        super(MultiSourceSignalGenerator, self).__init__(
            src_hypo_group_manager, dataset_list, data_list)

    def _construct_signal_candidates(self):
        """Constructs an array holding pointer information of signal candidate
        events pointing into the real MC dataset(s).
        """
        n_datasets = len(self._dataset_list)
        n_sources = self._src_hypo_group_manager.n_sources
        shg_list = self._src_hypo_group_manager.src_hypo_group_list
        sig_candidates_dtype = [
            ('ds_idx', get_smallest_numpy_int_type((0, n_datasets))),
            ('ev_idx', get_smallest_numpy_int_type(
                [0]+[len(data.mc) for data in self._data_list])),
            ('shg_idx', get_smallest_numpy_int_type((0, n_sources))),
            ('shg_src_idx', get_smallest_numpy_int_type(
                [0]+[shg.n_sources for shg in shg_list])),
            ('weight', np.float64)
        ]
        self._sig_candidates = np.empty(
            (0,), dtype=sig_candidates_dtype, order='F')

        # Go through the source hypothesis groups to get the signal event
        # candidates.
        for ((shg_idx, shg), (j, (ds, data))) in itertools.product(
                enumerate(shg_list),
                enumerate(zip(self._dataset_list, self._data_list))):
            sig_gen_method = shg.sig_gen_method
            if(sig_gen_method is None):
                raise ValueError(
                    'No signal generation method has been specified '
                    'for the %dth source hypothesis group!' % (shg_idx))
            data_mc = data.mc
            (ev_indices, src_indices, flux) = sig_gen_method.calc_source_signal_mc_event_flux(
                data_mc, shg)

            sig_candidates = np.empty(
                    (len(ev_indices),), dtype=sig_candidates_dtype, order='F'
                )
            sig_candidates['ds_idx'] = j
            sig_candidates['ev_idx'] = ev_indices
            sig_candidates['shg_idx'] = shg_idx
            sig_candidates['shg_src_idx'] = src_indices
            sig_candidates['weight'] = data_mc[ev_indices]['mcweight'] * data.livetime * 86400 * flux

            self._sig_candidates = np.append(self._sig_candidates, sig_candidates)
            del sig_candidates

        # Normalize the signal candidate weights.
        self._sig_candidates_weight_sum = np.sum(self._sig_candidates['weight'])
        self._sig_candidates['weight'] /= self._sig_candidates_weight_sum



# -*- coding: utf-8 -*-
# Author: Martin Wolf <mail@martin-wolf.org>

"""This test module tests classes, methods and functions of the ``core.model``
module.
"""

import unittest

from skyllh.core.model import (
    Model,
    ModelCollection
)


class Model_TestCase(unittest.TestCase):
    def setUp(self):
        self.model = Model('Model1')

    def test_name(self):
        self.assertEqual(self.model.name, 'Model1')

    def test_id(self):
        self.assertTrue(isinstance(self.model.id, int))


class ModelCollection_TestCase(unittest.TestCase):
    def setUp(self):
        self.model1 = Model('Model1')
        self.model2 = Model('Model2')
        self.modelcoll = ModelCollection((self.model1, self.model2))

    def test_cast(self):
        # Test cast function for None type.
        modelcoll = ModelCollection.cast(None)
        self.assertTrue(issubclass(modelcoll.model_type, Model))
        self.assertEqual(len(modelcoll.models), 0)

        # Test cast function for single Model instance.
        modelcoll = ModelCollection.cast(self.model1)
        self.assertTrue(issubclass(modelcoll.model_type, Model))
        self.assertEqual(len(modelcoll.models), 1)
        self.assertEqual(modelcoll.models[0], self.model1)

        # Test cast function for ModelCollection.
        modelcoll = ModelCollection.cast(self.modelcoll)
        self.assertEqual(modelcoll, self.modelcoll)

        # Test sequence of Model instances.
        modelcoll = ModelCollection.cast((self.model1, self.model2))
        self.assertTrue(issubclass(modelcoll.model_type, Model))
        self.assertEqual(len(modelcoll.models), 2)
        self.assertEqual(modelcoll.models[0], self.model1)
        self.assertEqual(modelcoll.models[1], self.model2)

        # Test that non-Model instances raises a TypeError.
        with self.assertRaises(TypeError):
            modelcoll = ModelCollection.cast('A str instance.')
        with self.assertRaises(TypeError):
            modelcoll = ModelCollection.cast(('str1','str2'))

    def test_model_type(self):
        self.assertTrue(issubclass(self.modelcoll.model_type, Model))

    def test_models(self):
        self.assertEqual(len(self.modelcoll.models), 2)
        self.assertEqual(self.modelcoll.models[0], self.model1)
        self.assertEqual(self.modelcoll.models[1], self.model2)


if(__name__ == '__main__'):
    unittest.main()



# -*- coding: utf-8 -*-

"""IceCube specific coordinate utility functions.
"""

import numpy as np

def azi_to_ra_transform(azi, mjd):
    """Rotates the given IceCube azimuth angles into right-ascention angles for
    the given MJD times. This function is IceCube specific and assumes that the
    detector is located excently at the South Pole and neglects all astronomical
    effects like Earth's precession.

    Parameters
    ----------
    azi : ndarray
        The array with the azimuth angles.
    mjd : ndarray
        The array with the MJD times for each azimuth angle.
    """
    _sidereal_length = 0.997269566 # sidereal day = length * solar day
    _sidereal_offset = 2.54199002505 # RA = offset + 2pi * (MJD/sidereal_length)%1 - azimuth
    sidereal_day_residuals = ((mjd/_sidereal_length)%1)
    ra = _sidereal_offset + 2 * np.pi * sidereal_day_residuals - azi
    ra = np.mod(ra, 2*np.pi)
    return ra

def ra_to_azi_transform(ra, mjd):
    """Rotates the given right-ascention angles to local IceCube azimuth angles.

    Parameters
    ----------
    ra : ndarray
        The array with the right-ascention angles.
    mjd : ndarray
        The array with the MJD times for each right-ascention angle.

    """
    # Use the azi_to_ra_transform function because it is symmetric.
    return azi_to_ra_transform(ra, mjd)

def hor_to_equ_transform(azi, zen, mjd):
    """Transforms the coordinate from the horizontal system (azimuth, zenith)
    into the equatorial system (right-ascention, declination) for detector at
    the South Pole and neglecting all astronomical effects like Earth
    precession.

    Parameters
    ----------
    azi : ndarray
        The azimuth angle.
    zen : ndarray
        The zenith angle.
    mjd : ndarray
        The time in MJD.

    Returns
    -------
    (ra, dec) : (ndarray, ndarray)
        The two-element tuple with the arrays of right-ascention and
        declination.
    """
    ra = azi_to_ra_transform(azi, mjd)
    dec = np.pi - zen
    return (ra, dec)



# -*- coding: utf-8 -*-

import numpy as np
import os.path
import unittest


from skyllh.core.parameters import (
    ParameterGrid,
)
from skyllh.core.signal_generator import (
    SignalGenerator,
)
from skyllh.core.source_hypothesis import (
    SourceHypoGroupManager,
)
from skyllh.core.source_hypo_group import (
    SourceHypoGroup,
)
from skyllh.i3.detsigyield import (
    PowerLawFluxPointLikeSourceI3DetSigYieldImplMethod,
)
from skyllh.i3.signal_generation import (
    PointLikeSourceI3SignalGenerationMethod,
)
from skyllh.physics.source import (
    PointLikeSource,
)
from skyllh.physics.flux import (
    PowerLawFlux,
)

DATA_SAMPLES_IMPORTED = True
try:
    from i3skyllh.datasets import repository
    from i3skyllh.datasets import data_samples
except:
    DATA_SAMPLES_IMPORTED = False


class TestSignalGenerator(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """This class method will run only once for this TestCase.
        """
        if not DATA_SAMPLES_IMPORTED:
            return

        dataset_name = 'PointSourceTracks_v004p00'
        dsc = data_samples[dataset_name].create_dataset_collection()
        ds_list = dsc.get_datasets(['IC86, 2018', 'IC86, 2019'])

        data_list = [ds.load_and_prepare_data() for ds in ds_list]

        sources = [
            PointLikeSource(ra=np.deg2rad(0), dec=np.deg2rad(10)),
            PointLikeSource(ra=np.deg2rad(30), dec=np.deg2rad(2)),
        ]

        fluxmodel = PowerLawFlux(
            Phi0=1,
            E0=1000,
            gamma=2)

        gamma_grid = ParameterGrid(name='gamma', grid=np.arange(1, 4.1, 0.1))
        detsigyield_builder = PowerLawFluxPointLikeSourceI3DetSigYieldImplMethod(
            gamma_grid=gamma_grid)

        sig_gen_method = PointLikeSourceI3SignalGenerationMethod()

        shg_mgr = SourceHypoGroupManager(
            SourceHypoGroup(
                sources=sources,
                fluxmodel=fluxmodel,
                detsigyield_implmethods=detsigyield_builder,
                sig_gen_method=sig_gen_method))

        cls._sig_gen = SignalGenerator(
            src_hypo_group_manager=shg_mgr,
            dataset_list=ds_list,
            data_list=data_list)

    @unittest.skipIf(not DATA_SAMPLES_IMPORTED, 'Data samples not imported!')
    def testSigCandidatesArray(self):

        arr = type(self)._sig_gen._sig_candidates

        # Check data type.
        self.assertTrue(
            isinstance(arr, np.ndarray)
        )

        # Check field names.
        field_names = arr.dtype.fields.keys()
        self.assertTrue(
            ('ds_idx' in field_names) and
            ('ev_idx' in field_names) and
            ('shg_idx' in field_names) and
            ('shg_src_idx' in field_names) and
            ('weight' in field_names)
        )

        # Check the length of the array.
        self.assertTrue(
            len(arr) == 894736,
            'array length'
        )
        # Check ds_idx values.
        ds_idxs = np.unique(arr['ds_idx'])
        self.assertTrue(
            ds_idxs.shape == (2,),
            'ds_idx shape'
        )
        self.assertTrue(
            np.all(np.equal(ds_idxs, np.array([0, 1]))),
            'ds_idx values'
        )

        # Check shg_idx values.
        shg_idxs = np.unique(arr['shg_idx'])
        self.assertTrue(
            shg_idxs.shape == (1,),
            'shg_idx shape'
        )
        self.assertTrue(
            np.all(np.equal(shg_idxs, np.array([0]))),
            'shg_idx values'
        )

        # Check shg_src_idx values.
        shg_src_idxs = np.unique(arr['shg_src_idx'])
        self.assertTrue(
            shg_src_idxs.shape == (2,),
            'shg_src_idxs shape'
        )
        self.assertTrue(
            np.all(np.equal(shg_src_idxs, np.array([0,1]))),
            'shg_idx values'
        )

    @unittest.skipIf(not DATA_SAMPLES_IMPORTED, 'Data samples not imported!')
    def testSigCandidatesWeightSum(self):
        self.assertTrue(
            type(self)._sig_gen._sig_candidates_weight_sum == 7884630181096259,
        )


if(__name__ == '__main__'):
    unittest.main()



# -*- coding: utf-8 -*-

import abc
import inspect
import numpy as np
import scipy.sparse

from skyllh.core.py import (
    classname,
    float_cast,
    issequenceof
)
from skyllh.core.source_hypothesis import SourceHypoGroupManager
from skyllh.core.timing import TaskTimer
from skyllh.physics.source import SourceModel


class EventSelectionMethod(object, metaclass=abc.ABCMeta):
    """This is the abstract base class for all event selection method classes.
    The idea is to pre-select only events that contribute to the likelihood
    function, i.e. are more signal than background like. The different methods
    are implemented through derived classes of this base class.
    """

    def __init__(self, src_hypo_group_manager):
        """Creates a new event selection method instance.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances.
        """
        super(EventSelectionMethod, self).__init__()

        self.src_hypo_group_manager = src_hypo_group_manager

        # The _src_arr variable holds a numpy record array with the necessary
        # source information needed for the event selection method.
        self._src_arr = self.source_to_array(
            self._src_hypo_group_manager.source_list)

    @property
    def src_hypo_group_manager(self):
        """The SourceHypoGroupManager instance, which defines the list of
        sources.
        """
        return self._src_hypo_group_manager
    @src_hypo_group_manager.setter
    def src_hypo_group_manager(self, manager):
        if(not isinstance(manager, SourceHypoGroupManager)):
            raise TypeError(
                'The src_hypo_group_manager property must be an instance of '
                'SourceHypoGroupManager!')
        self._src_hypo_group_manager = manager

    def change_source_hypo_group_manager(self, src_hypo_group_manager):
        """Changes the SourceHypoGroupManager instance of the event selection
        method. This will also recreate the internal source numpy record array.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The new SourceHypoGroupManager instance, that should be used for
            this event selection method.
        """
        self.src_hypo_group_manager = src_hypo_group_manager
        self._src_arr = self.source_to_array(self._src_hypo_group_manager.source_list)

    @abc.abstractmethod
    def source_to_array(self, sources):
        """This method is supposed to convert a sequence of SourceModel
        instances into a structured numpy ndarray with the source information
        in a format that is best understood by the actual event selection
        method.

        Parameters
        ----------
        sources : sequence of SourceModel
            The sequence of source models containing the necessary information
            of the source.

        Returns
        -------
        arr : numpy record ndarray
            The generated numpy record ndarray holding the necessary information
            for each source.
        """
        pass

    @abc.abstractmethod
    def select_events(self, events, ret_src_ev_idxs=False, tl=None):
        """This method selects the events, which will contribute to the
        log-likelihood ratio function.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the events.
        ret_src_ev_idxs : bool
            Flag if also the indices of the selected events should get
            returned as a (src_idxs, ev_idxs) tuple of 1d ndarrays.
            Default is False.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        selected_events : DataFieldRecordArray
            The instance of DataFieldRecordArray holding the selected events,
            i.e. a subset of the `events` argument.
        (src_idxs, ev_idxs) : 1d ndarrays of ints | None
            The indices of sources and selected events, in case
            `ret_src_ev_idxs` is set to True. Returns None, in case
            `ret_src_ev_idxs` is set to False.
        """
        pass


class AllEventSelectionMethod(EventSelectionMethod):
    """This event selection method selects all events.
    """
    def __init__(self, src_hypo_group_manager):
        """Creates a new event selection method instance.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances. For this particular
            event selection method it has no meaning, but it is an interface
            parameter.
        """
        super(AllEventSelectionMethod, self).__init__(
            src_hypo_group_manager)

    def source_to_array(self, sources):
        return None

    def select_events(self, events, ret_src_ev_idxs=False, tl=None):
        """Selects all of the given events. Hence, the returned event array is
        the same as the given array.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the events, for which
            the selection method should get applied.
        ret_src_ev_idxs : bool
            Flag if also the indices of the selected events should get
            returned as a (src_idxs, ev_idxs) tuple of 1d ndarrays.
            Default is False.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        selected_events : DataFieldRecordArray
            The instance of DataFieldRecordArray holding the selected events,
            i.e. a subset of the `events` argument.
        (src_idxs, ev_idxs) : 1d ndarrays of ints | None
            The indices of sources and selected events, in case
            `ret_src_ev_idxs` is set to True. Returns None, in case
            `ret_src_ev_idxs` is set to False.
        """
        if(ret_src_ev_idxs):
            # Calculate events indices.
            with TaskTimer(tl, 'ESM: Calculate indices of selected events.'):
                n_sources = self.src_hypo_group_manager.n_sources
                src_idxs = np.repeat(np.arange(n_sources), len(events.indices))
                ev_idxs = np.tile(events.indices, n_sources)

            return (events, (src_idxs, ev_idxs))

        return (events, None)


class SpatialEventSelectionMethod(EventSelectionMethod, metaclass=abc.ABCMeta):
    """This class defines the base class for all spatial event selection
    methods.
    """

    def __init__(self, src_hypo_group_manager):
        """Creates a new event selection method instance.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances.
        """
        super(SpatialEventSelectionMethod, self).__init__(
            src_hypo_group_manager)

    def source_to_array(self, sources):
        """Converts the given sequence of SourceModel instances into a
        structured numpy ndarray holding the necessary source information needed
        for this event selection method.

        Parameters
        ----------
        sources : sequence of SourceModel
            The sequence of source models containing the necessary information
            of the source.

        Returns
        -------
        arr : numpy record ndarray
            The generated numpy record ndarray holding the necessary information
            for each source. It contains the following data fields: 'ra', 'dec'.
        """
        if(not issequenceof(sources, SourceModel)):
            raise TypeError('The sources argument must be a sequence of '
                'SourceModel instances!')

        arr = np.empty(
            (len(sources),),
            dtype=[('ra', np.float64), ('dec', np.float64)],
            order='F')

        for (i, src) in enumerate(sources):
            arr['ra'][i] = src.loc.ra
            arr['dec'][i] = src.loc.dec

        return arr


class DecBandEventSectionMethod(SpatialEventSelectionMethod):
    """This event selection method selects events within a declination band
    around a list of point-like source positions.
    """
    def __init__(self, src_hypo_group_manager, delta_angle):
        """Creates and configures a spatial declination band event selection
        method object.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances.
        delta_angle : float
            The half-opening angle around the source in declination for which
            events should get selected.
        """
        super(DecBandEventSectionMethod, self).__init__(
            src_hypo_group_manager)

        self.delta_angle = delta_angle

    @property
    def delta_angle(self):
        """The half-opening angle around the source in declination and
        right-ascention for which events should get selected.
        """
        return self._delta_angle
    @delta_angle.setter
    def delta_angle(self, angle):
        angle = float_cast(angle, 'The delta_angle property must be castable '
            'to type float!')
        self._delta_angle = angle

    def select_events(
            self, events, ret_src_ev_idxs=False,
            ret_mask_idxs=False, tl=None):
        """Selects the events within the declination band.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray that holds the event data.
            The following data fields must exist:

            - 'dec' : float
                The declination of the event.
        ret_src_ev_idxs : bool
            Flag if also the indices of the selected events should get
            returned as a (src_idxs, ev_idxs) tuple of 1d ndarrays.
            Default is False.
        ret_mask_idxs : bool
            Flag if also the indices of the selected events mask should get
            returned as a mask_idxs 1d ndarray.
            Default is False.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        selected_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding only the selected
            events.
        idxs: where idxs is one of the following:
            - (src_idxs, ev_idxs) : 1d ndarrays of ints
                The indices of sources and selected events, in case
                `ret_src_ev_idxs` is set to True.
            - mask_idxs : 1d ndarrays of ints
                The indices of selected events mask, in case
                `ret_mask_idxs` is set to True.
            - None
                In case both `ret_src_ev_idxs` and `ret_mask_idxs` are set to
                False.
        """
        delta_angle = self._delta_angle
        src_arr = self._src_arr

        # Calculates the minus and plus declination around each source and
        # bound it to -90deg and +90deg, respectively.
        src_dec_minus = np.maximum(-np.pi/2, src_arr['dec'] - delta_angle)
        src_dec_plus = np.minimum(src_arr['dec'] + delta_angle, np.pi/2)

        # Determine the mask for the events which fall inside the declination
        # window.
        # mask_dec is a (N_sources,N_events)-shaped ndarray.
        with TaskTimer(tl, 'ESM-DecBand: Calculate mask_dec.'):
            mask_dec = ((events['dec'] > src_dec_minus[:,np.newaxis]) &
                        (events['dec'] < src_dec_plus[:,np.newaxis]))

        # Determine the mask for the events that fall inside at least one
        # source declination band.
        # mask is a (N_events,)-shaped ndarray.
        with TaskTimer(tl, 'ESM-DecBand: Calculate mask.'):
            mask = np.any(mask_dec, axis=0)

        # Reduce the events according to the mask.
        with TaskTimer(tl, 'ESM-DecBand: Create selected_events.'):
            # Using an integer indices array for data selection is several
            # factors faster than using a boolean array.
            mask_idxs = events.indices[mask]
            selected_events = events[mask_idxs]

        if(ret_src_ev_idxs and ret_mask_idxs):
            raise ValueError(
                'Only one of `ret_src_ev_idxs` and `ret_mask_idxs` can be set '
                'to True.')
        elif(ret_src_ev_idxs):
            # Get selected events indices.
            idxs = np.argwhere(mask_dec[:, mask])
            src_idxs = idxs[:, 0]
            ev_idxs = idxs[:, 1]
            return (selected_events, (src_idxs, ev_idxs))
        elif(ret_mask_idxs):
            return (selected_events, mask_idxs)

        return (selected_events, None)


class RABandEventSectionMethod(SpatialEventSelectionMethod):
    """This event selection method selects events within a right-ascension band
    around a list of point-like source positions.
    """
    def __init__(self, src_hypo_group_manager, delta_angle):
        """Creates and configures a right-ascension band event selection
        method object.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances.
        delta_angle : float
            The half-opening angle around the source in right-ascension for
            which events should get selected.
        """
        super(RABandEventSectionMethod, self).__init__(
            src_hypo_group_manager)

        self.delta_angle = delta_angle

    @property
    def delta_angle(self):
        """The half-opening angle around the source in declination and
        right-ascention for which events should get selected.
        """
        return self._delta_angle
    @delta_angle.setter
    def delta_angle(self, angle):
        angle = float_cast(angle,
            'The delta_angle property must be castable to type float!')
        self._delta_angle = angle

    def select_events(self, events, ret_src_ev_idxs=False, tl=None):
        """Selects the events within the right-ascention band.

        The solid angle dOmega = dRA * dSinDec = dRA * dDec * cos(dec) is a
        function of declination, i.e. for a constant dOmega, the right-ascension
        value has to change with declination.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray that holds the event data.
            The following data fields must exist:

            - 'ra' : float
                The right-ascention of the event.
            - 'dec' : float
                The declination of the event.
        ret_src_ev_idxs : bool
            Flag if also the indices of the selected events should get
            returned as a (src_idxs, ev_idxs) tuple of 1d ndarrays.
            Default is False.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        selected_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding only the selected
            events.
        (src_idxs, ev_idxs) : 1d ndarrays of ints | None
            The indices of sources and selected events, in case
            `ret_src_ev_idxs` is set to True. Returns None, in case
            `ret_src_ev_idxs` is set to False.
        """
        delta_angle = self._delta_angle
        src_arr = self._src_arr

        # Get the minus and plus declination around the sources.
        src_dec_minus = np.maximum(-np.pi/2, src_arr['dec'] - delta_angle)
        src_dec_plus = np.minimum(src_arr['dec'] + delta_angle, np.pi/2)

        # Calculate the cosine factor for the largest declination distance from
        # the source. We use np.amin here because smaller cosine values are
        # larger angles.
        # cosfact is a (N_sources,)-shaped ndarray.
        cosfact = np.amin(np.cos([src_dec_minus, src_dec_plus]), axis=0)

        # Calculate delta RA, which is a function of declination.
        # dRA is a (N_sources,)-shaped ndarray.
        dRA_half = np.amin(
            [np.repeat(2*np.pi, len(src_arr['ra'])),
             np.fabs(delta_angle / cosfact)], axis=0)

        # Calculate the right-ascension distance of the events w.r.t. the
        # source. We make sure to use the smaller distance on the circle, thus
        # the maximal distance is 180deg, i.e. pi.
        # ra_dist is a (N_sources,N_events)-shaped 2D ndarray.
        with TaskTimer(tl, 'ESM-RaBand: Calculate ra_dist.'):
            ra_dist = np.fabs(
                np.mod(events['ra'] - src_arr['ra'][:,np.newaxis] + np.pi, 2*np.pi) - np.pi)

        # Determine the mask for the events which fall inside the
        # right-ascention window.
        # mask_ra is a (N_sources,N_events)-shaped ndarray.
        with TaskTimer(tl, 'ESM-RaBand: Calculate mask_ra.'):
            mask_ra = ra_dist < dRA_half[:,np.newaxis]

        # Determine the mask for the events that fall inside at least one
        # source sky window.
        # mask is a (N_events,)-shaped ndarray.
        with TaskTimer(tl, 'ESM-RaBand: Calculate mask.'):
            mask = np.any(mask_ra, axis=0)

        # Reduce the events according to the mask.
        with TaskTimer(tl, 'ESM-RaBand: Create selected_events.'):
            # Using an integer indices array for data selection is several
            # factors faster than using a boolean array.
            selected_events = events[events.indices[mask]]

        if(ret_src_ev_idxs):
            # Get selected events indices.
            idxs = np.argwhere(mask_ra[:, mask])
            src_idxs = idxs[:, 0]
            ev_idxs = idxs[:, 1]

            return (selected_events, (src_idxs, ev_idxs))

        return (selected_events, None)


class SpatialBoxEventSelectionMethod(SpatialEventSelectionMethod):
    """This event selection method selects events within a spatial box in
    right-ascention and declination around a list of point-like source
    positions.
    """
    def __init__(self, src_hypo_group_manager, delta_angle):
        """Creates and configures a spatial box event selection method object.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances.
        delta_angle : float
            The half-opening angle around the source for which events should
            get selected.
        """
        super(SpatialBoxEventSelectionMethod, self).__init__(
            src_hypo_group_manager)

        self.delta_angle = delta_angle

    @property
    def delta_angle(self):
        """The half-opening angle around the source in declination and
        right-ascention for which events should get selected.
        """
        return self._delta_angle
    @delta_angle.setter
    def delta_angle(self, angle):
        angle = float_cast(angle,
            'The delta_angle property must be castable to type float!')
        self._delta_angle = angle

    def select_events(self, events, ret_src_ev_idxs=False, tl=None):
        """Selects the events within the spatial box in right-ascention and
        declination.

        The solid angle dOmega = dRA * dSinDec = dRA * dDec * cos(dec) is a
        function of declination, i.e. for a constant dOmega, the right-ascension
        value has to change with declination.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray that holds the event data.
            The following data fields must exist:

            - 'ra' : float
                The right-ascention of the event.
            - 'dec' : float
                The declination of the event.
        ret_src_ev_idxs : bool
            Flag if also the indices of the selected events should get
            returned as a (src_idxs, ev_idxs) tuple of 1d ndarrays.
            Default is False.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        selected_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding only the selected
            events.
        (src_idxs, ev_idxs) : 1d ndarrays of ints | None
            The indices of sources and selected events, in case
            `ret_src_ev_idxs` is set to True. Returns None, in case
            `ret_src_ev_idxs` is set to False.
        """
        delta_angle = self._delta_angle
        src_arr = self._src_arr

        # Get the minus and plus declination around the sources.
        src_dec_minus = np.maximum(-np.pi/2, src_arr['dec'] - delta_angle)
        src_dec_plus = np.minimum(src_arr['dec'] + delta_angle, np.pi/2)

        # Calculate the cosine factor for the largest declination distance from
        # the source. We use np.amin here because smaller cosine values are
        # larger angles.
        # cosfact is a (N_sources,)-shaped ndarray.
        cosfact = np.amin(np.cos([src_dec_minus, src_dec_plus]), axis=0)

        # Calculate delta RA, which is a function of declination.
        # dRA is a (N_sources,)-shaped ndarray.
        dRA_half = np.amin(
            [np.repeat(2*np.pi, len(src_arr['ra'])),
             np.fabs(delta_angle / cosfact)], axis=0)

        # Determine the mask for the events which fall inside the
        # right-ascention window.
        # mask_ra is a (N_sources,N_events)-shaped ndarray.
        with TaskTimer(tl, 'ESM: Calculate mask_ra.'):
            nsrc = len(src_arr['ra'])
            # Fill in batch sizes of 200 maximum to save memory.
            batch_size=200
            if nsrc > batch_size:
                mask_ra = np.zeros((nsrc, len(events['ra'])), dtype=bool)
                n_batches = int(np.ceil(nsrc / float(batch_size)))
                for bi in range(n_batches):
                    if not (bi == n_batches-1):
                        mask_ra[bi*batch_size : (bi+1)*batch_size,...] = (np.fabs(
                            np.mod(events['ra'] - src_arr['ra'][bi*batch_size : (bi+1)*batch_size][:,np.newaxis] + np.pi, 2*np.pi) -
                            np.pi) < dRA_half[ bi*batch_size : (bi+1)*batch_size ][:,np.newaxis])
                    else:
                        mask_ra[bi*batch_size : ,...] = (np.fabs(
                            np.mod(events['ra'] - src_arr['ra'][bi*batch_size:][:,np.newaxis] + np.pi, 2*np.pi) -
                            np.pi) < dRA_half[bi*batch_size:][:,np.newaxis])

            else:
                mask_ra = np.fabs(
                    np.mod(events['ra'] - src_arr['ra'][:,np.newaxis] + np.pi, 2*np.pi) - np.pi) < dRA_half[:,np.newaxis]

        # Determine the mask for the events which fall inside the declination
        # window.
        # mask_dec is a (N_sources,N_events)-shaped ndarray.
        with TaskTimer(tl, 'ESM: Calculate mask_dec.'):
            mask_dec = ((events['dec'] > src_dec_minus[:,np.newaxis]) &
                        (events['dec'] < src_dec_plus[:,np.newaxis]))

        # Determine the mask for the events which fall inside the
        # right-ascension and declination window.
        # mask_sky is a (N_sources,N_events)-shaped ndarray.
        with TaskTimer(tl, 'ESM: Calculate mask_sky.'):
            mask_sky = mask_ra & mask_dec
            del mask_ra
            del mask_dec

        # Determine the mask for the events that fall inside at least one
        # source sky window.
        # mask is a (N_events,)-shaped ndarray.
        with TaskTimer(tl, 'ESM: Calculate mask.'):
            mask = np.any(mask_sky, axis=0)

        # Reduce the events according to the mask.
        with TaskTimer(tl, 'ESM: Create selected_events.'):
            # Using an integer indices array for data selection is several
            # factors faster than using a boolean array.
            selected_events = events[events.indices[mask]]

        if(ret_src_ev_idxs):
            # Get selected events indices.
            idxs = np.argwhere(mask_sky[:, mask])
            src_idxs = idxs[:, 0]
            ev_idxs = idxs[:, 1]

            return (selected_events, (src_idxs, ev_idxs))

        return (selected_events, None)


class PsiFuncEventSelectionMethod(EventSelectionMethod):
    """This event selection method selects events whose psi value, i.e. the
    great circle distance of the event to the source, is smaller than the value
    of the provided function.
    """
    def __init__(self, src_hypo_group_manager, psi_name, func, axis_name_list):
        """Creates a new PsiFuncEventSelectionMethod instance.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances.
        psi_name : str
            The name of the data field that provides the psi value of the event.
        func : callable
            The function that should get evaluated for each event. The call
            signature must be ``func(*axis_data)``, where ``*axis_data`` is the
            event data of each required axis. The number of axes must match the
            provided axis names through the ``axis_name_list``.
        axis_name_list : list of str
            The list of data field names for each axis of the function ``func``.
            All field names must be valid field names of the trial data's
            DataFieldRecordArray instance.
        """
        super(PsiFuncEventSelectionMethod, self).__init__(
            src_hypo_group_manager)

        self.psi_name = psi_name
        self.func = func
        self.axis_name_list = axis_name_list

        if(not (len(inspect.signature(self._func).parameters) >=
                len(self._axis_name_list))):
            raise TypeError(
                'The func argument must be a callable instance with at least '
                '%d arguments!'%(
                    len(self._axis_name_list)))

        n_sources = self.src_hypo_group_manager.n_sources
        if(n_sources != 1):
            raise ValueError(
                'The `PsiFuncEventSelectionMethod.select_events` currently '
                f'supports only one source. It was called with {n_sources} '
                'sources.'
            )

    @property
    def psi_name(self):
        """The name of the data field that provides the psi value of the event.
        """
        return self._psi_name
    @psi_name.setter
    def psi_name(self, name):
        if(not isinstance(name, str)):
            raise TypeError(
                'The psi_name property must be an instance of type str!')
        self._psi_name = name

    @property
    def func(self):
        """The function that should get evaluated for each event. The call
        signature must be ``func(*axis_data)``, where ``*axis_data`` is the
        event data of each required axis. The number of axes must match the
        provided axis names through the ``axis_name_list`` property.
        """
        return self._func
    @func.setter
    def func(self, f):
        if(not callable(f)):
            raise TypeError(
                'The func property must be a callable instance!')
        self._func = f

    @property
    def axis_name_list(self):
        """The list of data field names for each axis of the function defined
        through the ``func`` property.
        """
        return self._axis_name_list
    @axis_name_list.setter
    def axis_name_list(self, names):
        if(not issequenceof(names, str)):
            raise TypeError(
                'The axis_name_list property must be a sequence of str '
                'instances!')
        self._axis_name_list = list(names)

    def source_to_array(self, sources):
        """Converts the given sequence of SourceModel instances into a
        structured numpy ndarray holding the necessary source information needed
        for this event selection method.

        Parameters
        ----------
        sources : sequence of SourceModel
            The sequence of source models containing the necessary information
            of the source.

        Returns
        -------
        arr : None
            Because this event selection method does not depend directly on the
            source (only indirectly through the psi values), no source array
            is required.
        """
        return None

    def select_events(self, events, ret_src_ev_idxs=False, tl=None):
        """Selects the events whose psi value is smaller than the value of the
        predefined function.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray that holds the event data.
            The following data fields must exist:

            - <psi_name> : float
                The great circle distance of the event with the source.
            - <*axis_name_list> : float
                The name of the axis required for the function ``func`` to be
                evaluated.

        ret_src_ev_idxs : bool
            Flag if also the indices of the selected events should get
            returned as a (src_idxs, ev_idxs) tuple of 1d ndarrays.
            Default is False.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        selected_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding only the selected
            events.
        (src_idxs, ev_idxs) : 1d ndarrays of ints | None
            The indices of sources and selected events, in case
            `ret_src_ev_idxs` is set to True. Returns None, in case
            `ret_src_ev_idxs` is set to False.
        """
        cls_name = classname(self)

        with TaskTimer(tl, '%s: Get psi values.'%(cls_name)):
            psi = events[self._psi_name]

        with TaskTimer(tl, '%s: Get axis data values.'%(cls_name)):
            func_args = [ events[axis] for axis in self._axis_name_list ]

        with TaskTimer(tl, '%s: Creating mask.'%(cls_name)):
            mask = psi < self._func(*func_args)

        with TaskTimer(tl, '%s: Create selected_events.'%(cls_name)):
            # Using an integer indices array for data selection is several
            # factors faster than using a boolean array.
            selected_events = events[events.indices[mask]]

        if(ret_src_ev_idxs):
            # Get selected events indices.
            idxs =  np.argwhere(np.atleast_2d(mask))
            src_idxs = idxs[:, 0]
            ev_idxs = idxs[:, 1]
            return (selected_events, (src_idxs, ev_idxs))

        return (selected_events, None)


class SpatialBoxAndPsiFuncEventSelectionMethod(SpatialBoxEventSelectionMethod):
    """This event selection method selects events within a spatial box in
    right-ascention and declination around a list of point-like source
    positions and performs an additional selection of events whose ang_err value
    is larger than the value of the provided function at a given psi value.
    """
    def __init__(self, src_hypo_group_manager, delta_angle, psi_name, func,
                 axis_name_list, psi_floor=None):
        """Creates and configures a spatial box and psi func event selection
        method object.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances.
        delta_angle : float
            The half-opening angle around the source for which events should
            get selected.
        psi_name : str
            The name of the data field that provides the psi value of the event.
        func : callable
            The function that should get evaluated for each event. The call
            signature must be ``func(*axis_data)``, where ``*axis_data`` is the
            event data of each required axis. The number of axes must match the
            provided axis names through the ``axis_name_list``.
        axis_name_list : list of str
            The list of data field names for each axis of the function ``func``.
            All field names must be valid field names of the trial data's
            DataFieldRecordArray instance.
        psi_floor : float | None
            The psi func event selection is excluded for events having psi value
            below the `psi_floor`. If None, set it to default 5 degrees.
        """
        super(SpatialBoxAndPsiFuncEventSelectionMethod, self).__init__(
            src_hypo_group_manager, delta_angle)

        self.psi_name = psi_name
        self.func = func
        self.axis_name_list = axis_name_list

        if(psi_floor is None):
            psi_floor = np.deg2rad(5)
        self.psi_floor = psi_floor

        if(not (len(inspect.signature(self._func).parameters) >=
                len(self._axis_name_list))):
            raise TypeError(
                'The func argument must be a callable instance with at least '
                '%d arguments!'%(
                    len(self._axis_name_list)))

    @property
    def psi_name(self):
        """The name of the data field that provides the psi value of the event.
        """
        return self._psi_name
    @psi_name.setter
    def psi_name(self, name):
        if(not isinstance(name, str)):
            raise TypeError(
                'The psi_name property must be an instance of type str!')
        self._psi_name = name

    @property
    def func(self):
        """The function that should get evaluated for each event. The call
        signature must be ``func(*axis_data)``, where ``*axis_data`` is the
        event data of each required axis. The number of axes must match the
        provided axis names through the ``axis_name_list`` property.
        """
        return self._func
    @func.setter
    def func(self, f):
        if(not callable(f)):
            raise TypeError(
                'The func property must be a callable instance!')
        self._func = f

    @property
    def axis_name_list(self):
        """The list of data field names for each axis of the function defined
        through the ``func`` property.
        """
        return self._axis_name_list
    @axis_name_list.setter
    def axis_name_list(self, names):
        if(not issequenceof(names, str)):
            raise TypeError(
                'The axis_name_list property must be a sequence of str '
                'instances!')
        self._axis_name_list = list(names)

    @property
    def psi_floor(self):
        """The psi func event selection is excluded for events having psi value
        below the `psi_floor`.
        """
        return self._psi_floor
    @psi_floor.setter
    def psi_floor(self, psi):
        psi = float_cast(psi, 'The psi_floor property must be castable '
            'to type float!')
        self._psi_floor = psi

    def _get_psi(self, events, idxs):
        """Function to calculate the the opening angle between the source
        position and the event's reconstructed position.
        """
        ra = events['ra']
        dec = events['dec']

        src_idxs, ev_idxs = idxs
        src_ra = self._src_arr['ra'][src_idxs]
        src_dec = self._src_arr['dec'][src_idxs]

        delta_dec = np.abs(np.take(dec, ev_idxs) - src_dec)
        delta_ra = np.abs(np.take(ra, ev_idxs) - src_ra)
        x = (np.sin(delta_dec / 2.))**2. + np.cos(np.take(dec, ev_idxs)) *\
            np.cos(src_dec) * (np.sin(delta_ra / 2.))**2.

        # Handle possible floating precision errors.
        x[x < 0.] = 0.
        x[x > 1.] = 1.

        psi = (2.0*np.arcsin(np.sqrt(x)))
        # Floor psi values below the first bin location in spatial KDE PDF.
        # Flooring at the boundary (1e-6) requires a regeneration of the
        # spatial KDE splines.
        floor = 10**(-5.95442953)
        psi = np.where(psi < floor, floor, psi)

        return psi

    def select_events(self, events, ret_src_ev_idxs=False, tl=None):
        """Selects the events within the spatial box in right-ascention and
        declination and performs an additional selection of events whose ang_err
        value is larger than the value of the provided function at a given psi
        value.

        The solid angle dOmega = dRA * dSinDec = dRA * dDec * cos(dec) is a
        function of declination, i.e. for a constant dOmega, the right-ascension
        value has to change with declination.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray that holds the event data.
            The following data fields must exist:

            - 'ra' : float
                The right-ascention of the event.
            - 'dec' : float
                The declination of the event.
        ret_src_ev_idxs : bool
            Flag if also the indices of the selected events should get
            returned as a (src_idxs, ev_idxs) tuple of 1d ndarrays.
            Default is False.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        selected_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding only the selected
            events.
        (src_idxs, ev_idxs) : 1d ndarrays of ints | None
            The indices of sources and selected events, in case
            `ret_src_ev_idxs` is set to True. Returns None, in case
            `ret_src_ev_idxs` is set to False.
        """
        delta_angle = self._delta_angle
        src_arr = self._src_arr

        # Get the minus and plus declination around the sources.
        src_dec_minus = np.maximum(-np.pi/2, src_arr['dec'] - delta_angle)
        src_dec_plus = np.minimum(src_arr['dec'] + delta_angle, np.pi/2)

        # Calculate the cosine factor for the largest declination distance from
        # the source. We use np.amin here because smaller cosine values are
        # larger angles.
        # cosfact is a (N_sources,)-shaped ndarray.
        cosfact = np.amin(np.cos([src_dec_minus, src_dec_plus]), axis=0)

        # Calculate delta RA, which is a function of declination.
        # dRA is a (N_sources,)-shaped ndarray.
        dRA_half = np.amin(
            [np.repeat(2*np.pi, len(src_arr['ra'])),
             np.fabs(delta_angle / cosfact)], axis=0)

        # Determine the mask for the events which fall inside the
        # right-ascention window.
        # mask_ra is a (N_sources,N_events)-shaped ndarray.
        with TaskTimer(tl, 'ESM: Calculate mask_ra.'):
            nsrc = len(src_arr['ra'])

            # Fill in batch sizes of 200 maximum to save memory.
            batch_size=200
            if nsrc > batch_size:
                mask_ra = np.zeros((nsrc, len(events['ra'])), dtype=bool)
                n_batches = int(np.ceil(nsrc / float(batch_size)))
                for bi in range(n_batches):
                    if not (bi == n_batches-1):
                        mask_ra[bi*batch_size : (bi+1)*batch_size,...] = (np.fabs(
                            np.mod(events['ra'] - src_arr['ra'][bi*batch_size : (bi+1)*batch_size][:,np.newaxis] + np.pi, 2*np.pi) -
                            np.pi) < dRA_half[ bi*batch_size : (bi+1)*batch_size ][:,np.newaxis])
                    else:
                        mask_ra[bi*batch_size : ,...] = (np.fabs(
                            np.mod(events['ra'] - src_arr['ra'][bi*batch_size:][:,np.newaxis] + np.pi, 2*np.pi) -
                            np.pi) < dRA_half[bi*batch_size:][:,np.newaxis])

            else:
                mask_ra = np.fabs(
                    np.mod(events['ra'] - src_arr['ra'][:,np.newaxis] + np.pi, 2*np.pi) - np.pi) < dRA_half[:,np.newaxis]

        # Determine the mask for the events which fall inside the declination
        # window.
        # mask_dec is a (N_sources,N_events)-shaped ndarray.
        with TaskTimer(tl, 'ESM: Calculate mask_dec.'):
            mask_dec = ((events['dec'] > src_dec_minus[:,np.newaxis]) &
                        (events['dec'] < src_dec_plus[:,np.newaxis]))

        # Determine the mask for the events which fall inside the
        # right-ascension and declination window.
        # mask_sky is a (N_sources,N_events)-shaped ndarray.
        with TaskTimer(tl, 'ESM: Calculate mask_sky.'):
            mask_sky = mask_ra & mask_dec
            del mask_ra
            del mask_dec

        # Determine the mask for the events that fall inside at least one
        # source sky window.
        # mask is a (N_events,)-shaped ndarray.
        with TaskTimer(tl, 'ESM: Calculate mask.'):
            mask = np.any(mask_sky, axis=0)

        # Reduce the events according to the mask.
        with TaskTimer(tl, 'ESM: Create selected_events.'):
            # Get selected events indices.
            idxs = np.argwhere(mask_sky[:, mask])
            src_idxs = idxs[:, 0]
            ev_idxs = idxs[:, 1]

            # Using an integer indices array for data selection is several
            # factors faster than using a boolean array.
            selected_events = events[events.indices[mask]]

        # Perform an additional selection based on psi values.
        with TaskTimer(tl, 'ESM: Get psi values.'):
            psi = self._get_psi(selected_events, (src_idxs, ev_idxs))

        with TaskTimer(tl, 'ESM: Create mask_psi.'):
            mask_psi = (
                (self._func(psi) <= selected_events['ang_err'][ev_idxs])
                | (psi < self.psi_floor)
            )

        with TaskTimer(tl, 'ESM: Create final_selected_events.'):
            # Have to define the shape argument in order to not truncate
            # the mask in case last events are not selected.
            final_mask_sky = scipy.sparse.csr_matrix(
                (mask_psi, (src_idxs, ev_idxs)),
                shape=(len(src_arr['ra']), len(selected_events))
            ).toarray()
            final_mask = np.any(final_mask_sky, axis=0)

            # Using an integer indices array for data selection is several
            # factors faster than using a boolean array.
            final_selected_events = selected_events[selected_events.indices[final_mask]]

        if(ret_src_ev_idxs):
            # Get final selected events indices.
            idxs = np.argwhere(final_mask_sky[:, final_mask])
            src_idxs = idxs[:, 0]
            ev_idxs = idxs[:, 1]

            return (final_selected_events, (src_idxs, ev_idxs))

        return (final_selected_events, None)



# -*- coding: utf-8 -*-

import abc
import copy
import pickle
import os.path
import numpy as np
import sys

from skyllh.core.py import (
    classname,
    get_byte_size_prefix,
    getsizeof,
    issequence,
    issequenceof
)
from skyllh.core import display as dsp


# Define a file loader registry that holds the FileLoader classes for different
# file formats.
_FILE_LOADER_REG = dict()

def register_FileLoader(formats, fileloader_cls):
    """Registers the given file formats (file extensions) to the given
    FileLoader class.

    Parameters
    ----------
    formats : str | list of str
        The list of file name extensions that should be mapped to the FileLoader
        class.
    fileloader_cls : FileLoader
        The subclass of FileLoader that should be used for the given file
        formats.
    """
    if(isinstance(formats, str)):
        formats = [ formats ]
    if(not issequence(formats)):
        raise TypeError('The "formats" argument must be a sequence!')
    if(not issubclass(fileloader_cls, FileLoader)):
        raise TypeError('The "fileloader_cls" argument must be a subclass of FileLoader!')

    for fmt in formats:
        if(fmt in _FILE_LOADER_REG.keys()):
            raise KeyError('The format "%s" is already registered!'%(fmt))
        _FILE_LOADER_REG[fmt] = fileloader_cls

def create_FileLoader(pathfilenames, **kwargs):
    """Creates the appropriate FileLoader object for the given file names.
    It looks up the FileLoader class from the FileLoader registry for the
    file name extension of the first file name in the given list.

    Parameters
    ----------
    pathfilenames : str | sequence of str
        The sequence of fully qualified file names of the files that should be
        loaded.

    Additional Parameters
    ---------------------
    Additional parameters will be passed to the constructor method of the
    chosen FileLoader class.

    Returns
    -------
    fileloader : FileLoader
        The appropiate FileLoader instance for the given type of data files.
    """
    if(isinstance(pathfilenames, str)):
        pathfilenames = [pathfilenames]
    if(not issequenceof(pathfilenames, str)):
        raise TypeError('The pathfilenames argument must be a sequence of str!')

    # Sort the file names extensions with shorter extensions before longer ones
    # to support a format that is sub-string of another format.
    formats = sorted(_FILE_LOADER_REG.keys())
    for fmt in formats:
        l = len(fmt)
        if(pathfilenames[0][-l:].lower() == fmt.lower()):
            cls = _FILE_LOADER_REG[fmt]
            return cls(pathfilenames, **kwargs)

    raise RuntimeError('No FileLoader class is suitable to load the data file "%s"!'%(pathfilenames[0]))

def assert_file_exists(pathfilename):
    """Checks if the given file exists and raises a RuntimeError if it does
    not exist.
    """
    if(not os.path.isfile(pathfilename)):
        raise RuntimeError('The data file "%s" does not exist!'%(pathfilename))


class FileLoader(object, metaclass=abc.ABCMeta):

    def __init__(self, pathfilenames, **kwargs):
        """Initializes a new FileLoader instance.

        Parameters
        ----------
        pathfilenames : str | sequence of str
            The sequence of fully qualified file names of the data files that
            need to be loaded.
        """
        super(FileLoader, self).__init__(**kwargs)

        self.pathfilename_list = pathfilenames

    @property
    def pathfilename_list(self):
        """The list of fully qualified file names of the data files.
        """
        return self._pathfilename_list
    @pathfilename_list.setter
    def pathfilename_list(self, pathfilenames):
        if(isinstance(pathfilenames, str)):
            pathfilenames = [ pathfilenames ]
        if(not issequence(pathfilenames)):
            raise TypeError('The pathfilename_list property must be of type '
                'str or a sequence of type str!')
        self._pathfilename_list = list(pathfilenames)

    @abc.abstractmethod
    def load_data(self, **kwargs):
        """This method is supposed to load the data from the file.
        """
        pass


class NPYFileLoader(FileLoader):
    """The NPYFileLoader class provides the data loading functionality for
    numpy data files containing numpy arrays. It uses the ``numpy.load``
    function for loading the data and the numpy.append function to concatenate
    several data files.
    """
    def __init__(self, pathfilenames, **kwargs):
        super(NPYFileLoader, self).__init__(pathfilenames)

    def _load_file_memory_efficiently(
            self, pathfilename, keep_fields, dtype_convertions,
            dtype_convertion_except_fields):
        """Loads a single file in a memory efficient way.

        Parameters
        ----------
        pathfilename : str
            The fully qualified file name of the to-be-loaded file.

        Returns
        -------
        data : DataFieldRecordArray instance
            An instance of DataFieldRecordArray holding the data.
        """
        assert_file_exists(pathfilename)
        # Create a memory map into the data file. This loads the data only when
        # accessing the data.
        mmap_ndarray = np.load(pathfilename, mmap_mode='r')
        field_names = mmap_ndarray.dtype.names
        fname_to_fidx = dict(
            [ (fname,idx) for (idx,fname) in enumerate(field_names) ])
        dt_fields = mmap_ndarray.dtype.fields
        n_rows = mmap_ndarray.shape[0]

        data = dict()

        # Create empty arrays for each column of length n_rows.
        for fname in field_names:
            # Ignore fields that should not get kept.
            if((keep_fields is not None) and (fname not in keep_fields)):
                continue

            # Get the original data type of the field.
            dt = dt_fields[fname][0]
            # Convert the data type if requested.
            if((fname not in dtype_convertion_except_fields) and
               (dt in dtype_convertions)):
                dt = dtype_convertions[dt]

            data[fname] = np.empty((n_rows,), dtype=dt)

        # Loop through the rows of the recarray.
        bs = 4096
        for ridx in range(0, n_rows):
            row = mmap_ndarray[ridx]
            for fname in data.keys():
                fidx = fname_to_fidx[fname]
                data[fname][ridx] = row[fidx]

            # Reopen the data file after each given blocksize.
            if(ridx % bs == 0):
                del mmap_ndarray
                mmap_ndarray = np.load(pathfilename, mmap_mode='r')

        # Close the memory map file.
        del mmap_ndarray

        # Create a DataFieldRecordArray out of the dictionary.
        data = DataFieldRecordArray(data, copy=False)

        return data

    def _load_file_time_efficiently(
            self, pathfilename, keep_fields, dtype_convertions,
            dtype_convertion_except_fields):
        """Loads a single file in a time efficient way. This will load the data
        column-wise.
        """
        assert_file_exists(pathfilename)
        # Create a memory map into the data file. This loads the data only when
        # accessing the data.
        mmap_ndarray = np.load(pathfilename, mmap_mode='r')

        # Create a DataFieldRecordArray out of the memory mapped file. We need
        # to copy the data, otherwise we get read-only numpy arrays.
        data = DataFieldRecordArray(
            mmap_ndarray,
            keep_fields=keep_fields,
            dtype_convertions=dtype_convertions,
            dtype_convertion_except_fields=dtype_convertion_except_fields,
            copy=True)

        # Close the memory map file.
        del mmap_ndarray

        return data

    def load_data(
            self, keep_fields=None, dtype_convertions=None,
            dtype_convertion_except_fields=None, efficiency_mode=None):
        """Loads the data from the files specified through their fully qualified
        file names.

        Parameters
        ----------
        keep_fields : str | sequence of str | None
            Load the data into memory only for these data fields. If set to
            ``None``, all in-file-present data fields are loaded into memory.
        dtype_convertions : dict | None
            If not None, this dictionary defines how data fields of specific
            data types get converted into the specified data types.
            This can be used to use less memory.
        dtype_convertion_except_fields : str | sequence of str | None
            The sequence of field names whose data type should not get
            converted.
        efficiency_mode : str | None
            The efficiency mode the data should get loaded with. Possible values
            are:

                - 'memory':
                    The data will be load in a memory efficient way. This will
                    require more time, because all data records of a file will
                    be loaded sequentially.
                - 'time'
                    The data will be loaded in a time efficient way. This will
                    require more memory, because each data file gets loaded in
                    memory at once.

            The default value is ``'time'``. If set to ``None``, the default
            value will be used.

        Returns
        -------
        data : DataFieldRecordArray
            The DataFieldRecordArray holding the loaded data.

        Raises
        ------
        RuntimeError if a file does not exist.
        """
        if(keep_fields is not None):
            if(isinstance(keep_fields, str)):
                keep_fields = [ keep_fields ]
            elif(not issequenceof(keep_fields, str)):
                raise TypeError('The keep_fields argument must be None, an '
                    'instance of type str, or a sequence of instances of '
                    'type str!')

        if(dtype_convertions is None):
            dtype_convertions = dict()
        elif(not isinstance(dtype_convertions, dict)):
            raise TypeError('The dtype_convertions argument must be None, '
                'or an instance of dict!')

        if(dtype_convertion_except_fields is None):
            dtype_convertion_except_fields = []
        elif(isinstance(dtype_convertion_except_fields, str)):
            dtype_convertion_except_fields = [ dtype_convertion_except_fields ]
        elif(not issequenceof(dtype_convertion_except_fields, str)):
            raise TypeError('The dtype_convertion_except_fields argument '
                'must be a sequence of str instances.')

        efficiency_mode2func = {
            'memory': self._load_file_memory_efficiently,
            'time': self._load_file_time_efficiently
        }
        if(efficiency_mode is None):
            efficiency_mode = 'time'
        if(not isinstance(efficiency_mode, str)):
            raise TypeError('The efficiency_mode argument must be an instance '
                'of type str!')
        if(efficiency_mode not in efficiency_mode2func):
            raise ValueError('The efficiency_mode argument value must be one '
                'of %s!'%(', '.join(efficiency_mode2func.keys())))
        load_file_func = efficiency_mode2func[efficiency_mode]

        # Load the first data file.
        data = load_file_func(
            self._pathfilename_list[0],
            keep_fields=keep_fields,
            dtype_convertions=dtype_convertions,
            dtype_convertion_except_fields=dtype_convertion_except_fields
        )

        # Load possible subsequent data files by appending to the first data.
        for i in range(1, len(self._pathfilename_list)):
            data.append(load_file_func(
                self._pathfilename_list[i],
                keep_fields=keep_fields,
                dtype_convertions=dtype_convertions,
                dtype_convertion_except_fields=dtype_convertion_except_fields
            ))

        return data


class PKLFileLoader(FileLoader):
    """The PKLFileLoader class provides the data loading functionality for
    pickled Python data files containing Python data structures. It uses the
    `pickle.load` function for loading the data from the file.
    """
    def __init__(self, pathfilenames, pkl_encoding=None, **kwargs):
        """Creates a new file loader instance for a pickled data file.

        Parameters
        ----------
        pathfilenames : str | sequence of str
            The sequence of fully qualified file names of the data files that
            need to be loaded.
        pkl_encoding : str | None
            The encoding of the pickled data files. If None, the default
            encodings 'ASCII' and 'latin1' will be tried to load the data.
        """
        super(PKLFileLoader, self).__init__(pathfilenames)

        self.pkl_encoding = pkl_encoding

    @property
    def pkl_encoding(self):
        """The encoding of the pickled data files. Can be None.
        If None, the default encodings 'ASCII' and 'latin1' will be tried to
        load the data.
        """
        return self._pkl_encoding
    @pkl_encoding.setter
    def pkl_encoding(self, encoding):
        if(encoding is not None):
            if(not isinstance(encoding, str)):
                raise TypeError('The pkl_encoding property must be None or of '
                    'type str!')
        self._pkl_encoding = encoding

    def load_data(self, **kwargs):
        """Loads the data from the files specified through their fully qualified
        file names.

        Returns
        -------
        data : Python object | list of Python objects
            The de-pickled Python object. If more than one file was specified,
            this is a list of Python objects, i.e. one object for each file.
            The file <-> object mapping order is preserved.

        Raises
        ------
        RuntimeError if a file does not exist.
        """
        # Define the possible encodings of the pickled files.
        encodings = ['ASCII', 'latin1']
        if(self._pkl_encoding is not None):
            encodings = [self._pkl_encoding] + encodings

        data = []
        for pathfilename in self.pathfilename_list:
            assert_file_exists(pathfilename)
            with open(pathfilename, 'rb') as ifile:
                enc_idx = 0
                load_ok = False
                obj = None
                while (not load_ok) and (enc_idx < len(encodings)):
                    try:
                        encoding = encodings[enc_idx]
                        obj = pickle.load(ifile, encoding=encoding)
                    except UnicodeDecodeError:
                        enc_idx += 1
                        # Move the file pointer back to the beginning of the
                        # file.
                        ifile.seek(0)
                    else:
                        load_ok = True
                if(obj is None):
                    raise RuntimeError('The file "%s" could not get unpickled! '
                        'No correct encoding available!'%(pathfilename))
                data.append(obj)

        if(len(data) == 1):
            data = data[0]

        return data


class DataFieldRecordArray(object):
    """The DataFieldRecordArray class provides a data container similar to a numpy
    record ndarray. But the data fields are stored as individual numpy ndarray
    objects. Hence, access of single data fields is much faster compared to
    access on the record ndarray.
    """
    def __init__(self, data, keep_fields=None, dtype_convertions=None,
            dtype_convertion_except_fields=None, copy=True):
        """Creates a DataFieldRecordArray from the given numpy record ndarray.

        Parameters
        ----------
        data : numpy record ndarray | dict | DataFieldRecordArray | None
            The numpy record ndarray that needs to get transformed into the
            DataFieldRecordArray instance. Alternative a dictionary with field
            names as keys and numpy ndarrays as values can be provided. If an
            instance of DataFieldRecordArray is provided, the new
            DataFieldRecordArray gets constructed from the copy of the data of
            the provided DataFieldRecordArray instance.
            If set to `None`, the DataFieldRecordArray instance is initialized
            with no data and the length of the array is set to 0.
        keep_fields : str | sequence of str | None
            If not None (default), this specifies the data fields that should
            get kept from the given data. Otherwise all data fields get kept.
        dtype_convertions : dict | None
            If not None, this dictionary defines how data fields of specific
            data types get converted into the specified data types.
            This can be used to use less memory.
        dtype_convertion_except_fields : str | sequence of str | None
            The sequence of field names whose data type should not get
            converted.
        copy : bool
            Flag if the input data should get copied. Default is True. If a
            DataFieldRecordArray instance is provided, this option is set to
            `True` automatically.
        """
        self._data_fields = dict()
        self._len = None

        if(data is None):
            data = dict()

        if(keep_fields is not None):
            if(isinstance(keep_fields, str)):
                keep_fields = [ keep_fields ]
            elif(not issequenceof(keep_fields, str)):
                raise TypeError('The keep_fields argument must be None, an '
                    'instance of type str, or a sequence of instances of '
                    'type str!')

        if(dtype_convertions is None):
            dtype_convertions = dict()
        elif(not isinstance(dtype_convertions, dict)):
            raise TypeError('The dtype_convertions argument must be None, '
                'or an instance of dict!')

        if(dtype_convertion_except_fields is None):
            dtype_convertion_except_fields = []
        elif(isinstance(dtype_convertion_except_fields, str)):
            dtype_convertion_except_fields = [ dtype_convertion_except_fields ]
        elif(not issequenceof(dtype_convertion_except_fields, str)):
            raise TypeError('The dtype_convertion_except_fields argument '
                'must be a sequence of str instances.')

        if(isinstance(data, np.ndarray)):
            field_names = data.dtype.names
            fname2dtype = dict(
                [(k,v[0]) for (k,v) in data.dtype.fields.items() ])
            length = data.shape[0]
        elif(isinstance(data, dict)):
            field_names = list(data.keys())
            fname2dtype = dict(
                [ (fname, data[fname].dtype) for fname in field_names ])
            length = 0
            if(len(field_names) > 0):
                length = data[field_names[0]].shape[0]
        elif(isinstance(data, DataFieldRecordArray)):
            field_names = data.field_name_list
            fname2dtype = dict(
                [ (fname, data[fname].dtype) for fname in field_names ])
            length = len(data)
            copy = True
        else:
            raise TypeError('The data argument must be an instance of ndarray, '
                'dict, or DataFieldRecordArray!')

        for fname in field_names:
            # Ignore fields that should not get kept.
            if((keep_fields is not None) and (fname not in keep_fields)):
                continue

            copy_field = copy
            dt = fname2dtype[fname]
            if((fname not in dtype_convertion_except_fields) and
               (dt in dtype_convertions)):
                dt = dtype_convertions[dt]
                # If a data type convertion is needed, the data of the field
                # needs to get copied.
                copy_field = True

            if(copy_field is True):
                # Create a ndarray with the final data type and then assign the
                # values from the data, which technically is a copy.
                field_arr = np.empty((length,), dtype=dt)
                field_arr[:] = data[fname]
            else:
                field_arr = data[fname]
            if(self._len is None):
                self._len = len(field_arr)
            elif(len(field_arr) != self._len):
                raise ValueError('All field arrays must have the same length. '
                    'Field "%s" has length %d, but must be %d!'%(
                    fname, len(field_arr), self._len))

            self._data_fields[fname] = field_arr

        if(self._len is None):
            # The DataFieldRecordArray is initialized with no fields, i.e. also
            # also no data.
            self._len = 0

        self._field_name_list = list(self._data_fields.keys())
        self._indices = None

    def __contains__(self, name):
        """Checks if the given field exists in this DataFieldRecordArray
        instance.

        Parameters
        ----------
        name : str
            The name of the field.

        Returns
        -------
        check : bool
            True, if the given field exists in this DataFieldRecordArray
            instance, False otherwise.
        """
        return (name in self._data_fields)

    def __getitem__(self, name):
        """Implements data field value access.

        Parameters
        ----------
        name : str | numpy ndarray of int or bool
            The name of the data field. If a numpy ndarray is given, it must
            contain the indices for which to retrieve a data selection of the
            entire DataFieldRecordArray. A numpy ndarray of bools can be given
            as well to define a mask.

        Raises
        ------
        KeyError
            If the given data field does not exist.

        Returns
        -------
        data : numpy ndarray | DataFieldRecordArray
            The requested field data or a DataFieldRecordArray holding the
            requested selection of the entire data.
        """
        if(isinstance(name, np.ndarray)):
            return self.get_selection(name)

        if(name not in self._data_fields):
            raise KeyError('The data field "%s" is not present in the '
                'DataFieldRecordArray instance.'%(name))

        return self._data_fields[name]

    def __setitem__(self, name, arr):
        """Implements data field value assigment. If values are assigned to a
        data field that does not exist yet, it  will be added via the
        ``append_field`` method.

        Parameters
        ----------
        name : str | numpy ndarray of int or bool
            The name of the data field, or a numpy ndarray holding the indices
            or mask of a selection of this DataFieldRecordArray.
        arr : numpy ndarray | instance of DataFieldRecordArray
            The numpy ndarray holding the field values. It must be of the same
            length as this DataFieldRecordArray. If `name` is a numpy ndarray,
            `arr` must be a DataFieldRecordArray.

        Raises
        ------
        ValueError
            If the given data array is not of the same length as this
            DataFieldRecordArray instance.
        """
        if(isinstance(name, np.ndarray)):
            self.set_selection(name, arr)
            return

        # Check if a new field is supposed to be added.
        if(name not in self):
            self.append_field(name, arr)
            return

        # We set a particular already existing data field.
        if(len(arr) != self._len):
            raise ValueError('The length of the to-be-set data (%d) must '
                'match the length (%d) of the DataFieldRecordArray instance!'%(
                len(arr), self._len))

        if(not isinstance(arr, np.ndarray)):
            raise TypeError(
                'When setting a field directly, the data must be provided as a '
                'numpy ndarray!')

        self._data_fields[name] = arr

    def __len__(self):
        return self._len

    def __sizeof__(self):
        """Calculates the size in bytes of this DataFieldRecordArray instance
        in memory.

        Returns
        -------
        memsize : int
            The memory size in bytes that this DataFieldRecordArray instance
            has.
        """
        memsize = getsizeof([
            self._data_fields,
            self._len,
            self._field_name_list,
            self._indices
        ])
        return memsize

    def __str__(self):
        """Creates a pretty informative string representation of this
        DataFieldRecordArray instance.
        """
        (size, prefix) = get_byte_size_prefix(sys.getsizeof(self))

        max_field_name_len = np.max(
            [len(fname) for fname in self._field_name_list])

        # Generates a pretty string representation of the given field name.
        def _pretty_str_field(name):
            field = self._data_fields[name]
            s = '%s: {dtype: %s, vmin: % .3e, vmax: % .3e}'%(
                name.ljust(max_field_name_len), str(field.dtype), np.min(field),
                np.max(field))
            return s

        indent_str = ' '*dsp.INDENTATION_WIDTH
        s = '%s: %d fields, %d entries, %.0f %sbytes '%(
            classname(self), len(self._field_name_list), self.__len__(),
            np.round(size, 0), prefix)
        if(len(self._field_name_list) > 0):
            s += '\n' + indent_str + 'fields = {'
            s += '\n' + indent_str*2 + _pretty_str_field(self._field_name_list[0])
            for fname in self._field_name_list[1:]:
                s += '\n' + indent_str*2 + _pretty_str_field(fname)
            s += '\n' + indent_str + '}'
        return s

    @property
    def field_name_list(self):
        """(read-only) The list of the field names of this DataFieldRecordArray.
        """
        return self._field_name_list

    @property
    def indices(self):
        """(read-only) The numpy ndarray holding the indices of this
        DataFieldRecordArray.
        """
        if(self._indices is None):
            self._indices = np.indices((self._len,))[0]
        return self._indices

    def append(self, arr):
        """Appends the given DataFieldRecordArray to this DataFieldRecordArray
        instance.

        Parameters
        ----------
        arr : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray that should get appended to
            this DataFieldRecordArray. It must contain the same data fields.
            Additional data fields are ignored.
        """
        if(not isinstance(arr, DataFieldRecordArray)):
            raise TypeError('The arr argument must be an instance of '
                'DataFieldRecordArray!')

        for fname in self._field_name_list:
            self._data_fields[fname] = np.append(
                self._data_fields[fname], arr[fname])

        self._len += len(arr)
        self._indices = None

    def append_field(self, name, data):
        """Appends a field and its data to this DataFieldRecordArray instance.

        Parameters
        ----------
        name : str
            The name of the new data field.
        data : numpy ndarray
            The numpy ndarray holding the data.

        Raises
        ------
        KeyError
            If the given data field name already exists in this
            DataFieldRecordArray instance.
        ValueError
            If the length of the data array does not equal to the length of the
            data of this DataFieldRecordArray instance.
        TypeError
            If the arguments are of the wrong type.
        """
        if(not isinstance(name, str)):
            raise TypeError(
                'The name argument must be an instance of str!')
        if(not isinstance(data, np.ndarray)):
            raise TypeError(
                'The data argument must be an instance of ndarray!')
        if(name in self._data_fields):
            raise KeyError(
                'The data field "%s" already exists in this %s instance!'%(
                    name, classname(self)))
        #if(len(data) != self._len):
        #    raise ValueError(
        #        'The length of the given data is %d, but must be %d!'%(
        #            len(data), self._len))

        self._data_fields[name] = data
        self._field_name_list.append(name)

    def as_numpy_record_array(self):
        """Creates a numpy record ndarray instance holding the data of this
        DataFieldRecordArray instance.

        Returns
        -------
        arr : numpy record ndarray
            The numpy recarray ndarray holding the data of this
            DataFieldRecordArray instance.
        """
        dt = np.dtype(
            [(name, self._data_fields[name].dtype)
             for name in self.field_name_list])

        arr = np.empty((len(self),), dtype=dt)
        for name in self.field_name_list:
            arr[name] = self[name]

        return arr

    def copy(self, keep_fields=None):
        """Creates a new DataFieldRecordArray that is a copy of this
        DataFieldRecordArray instance.

        Parameters
        ----------
        keep_fields : str | sequence of str | None
            If not None (default), this specifies the data fields that should
            get kept from this DataFieldRecordArray. Otherwise all data fields
            get kept.
        """
        return DataFieldRecordArray(self, keep_fields=keep_fields)

    def remove_field(self, name):
        """Removes the given field from this array.

        Parameters
        ----------
        name : str
            The name of the data field that is to be removed.
        """
        self._data_fields.pop(name)
        self._field_name_list.remove(name)

    def get_field_dtype(self, name):
        """Returns the numpy dtype object of the given data field.
        """
        return self._data_fields[name].dtype

    def set_field_dtype(self, name, dt):
        """Sets the data type of the given field.

        Parameters
        ----------
        name : str
            The name of the data field.
        dt : numpy.dtype
            The dtype instance defining the new data type.
        """
        if(name not in self):
            raise KeyError(
                f'The data field "{name}" does not exist in this '
                 'DataFieldRecordArray!')
        if(not isinstance(dt, np.dtype)):
            raise TypeError(
                'The dt argument must be an instance of type numpy.dtype!')

        self._data_fields[name] = self._data_fields[name].astype(dt, copy=False)

    def convert_dtypes(self, convertions, except_fields=None):
        """Converts the data type of the data fields of this
        DataFieldRecordArray. This method can be used to compress the data.

        Parameters
        ----------
        convertions : dict of `old_dtype` -> `new_dtype`
            The dictionary with the old dtype as key and the new dtype as value.
        except_fields : sequence of str | None
            The sequence of field names, which should not get converted.
        """
        if(not isinstance(convertions, dict)):
            raise TypeError('The convertions argument must be an instance of '
                'dict!')

        if(except_fields is None):
            except_fields = []
        if(not issequenceof(except_fields, str)):
            raise TypeError('The except_fields argument must be a sequence '
                'of str!')

        _data_fields = self._data_fields
        for fname in self._field_name_list:
            if(fname in except_fields):
                continue
            old_dtype = _data_fields[fname].dtype
            if(old_dtype in convertions):
                new_dtype = convertions[old_dtype]
                _data_fields[fname] = _data_fields[fname].astype(new_dtype)

    def get_selection(self, indices):
        """Creates an DataFieldRecordArray that contains a selection of the data
        of this DataFieldRecordArray instance.

        Parameters
        ----------
        indices : (N,)-shaped numpy ndarray of int or bool
            The numpy ndarray holding the indices for which to select the data.

        Returns
        -------
        data_field_array : DataFieldRecordArray
            The DataFieldRecordArray that contains the selection of the
            original DataFieldRecordArray. The selection data is a copy of the
            original data.
        """
        data = dict()
        for fname in self._field_name_list:
            # Get the data selection from the original data. This creates a
            # copy.
            data[fname] = self._data_fields[fname][indices]
        return DataFieldRecordArray(data, copy=False)

    def set_selection(self, indices, arr):
        """Sets a selection of the data of this DataFieldRecordArray instance
        to the data given in arr.

        Parameters
        ----------
        indices : (N,)-shaped numpy ndarray of int or bool
            The numpy ndarray holding the indices or mask for which to set the
            data.
        arr : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the selection data.
            It must have the same fields defined as this DataFieldRecordArray
            instance.
        """
        if(not isinstance(arr, DataFieldRecordArray)):
            raise TypeError('The arr argument must be an instance of '
                'DataFieldRecordArray!')

        for fname in self._field_name_list:
            self._data_fields[fname][indices] = arr[fname]

    def rename_fields(self, convertions, must_exist=False):
        """Renames the given fields of this array.

        Parameters
        ----------
        convertions : dict of `old_name` -> `new_name`
            The dictionary holding the old and new names of the data fields.
        must_exist : bool
            Flag if the given fields must exist. If set to ``True`` and a field
            does not exist, a KeyError is raised.

        Raises
        ------
        KeyError
            If ``must_exist`` is set to ``True`` and a given field does not
            exist.
        """
        for (old_fname, new_fname) in convertions.items():
            if(old_fname in self.field_name_list):
                self._data_fields[new_fname] = self._data_fields.pop(old_fname)
            elif(must_exist is True):
                raise KeyError('The required field "%s" does not exist!'%(
                    old_fname))

        self._field_name_list = list(self._data_fields.keys())

    def tidy_up(self, keep_fields):
        """Removes all fields that are not specified through the keep_fields
        argument.

        Parameters
        ----------
        keep_fields : str | sequence of str
            The field name(s), that should not be removed.

        Raises
        ------
        TypeError
            If keep_fields is not an instance of str or a sequence of str
            instances.
        """
        if(isinstance(keep_fields, str)):
            keep_fields = [ keep_fields ]
        if(not issequenceof(keep_fields, str)):
            raise TypeError('The keep_fields argument must be a sequence of '
                'str!')

        # We need to make a copy of the field_name_list because that list will
        # get changed by the `remove_field` method.
        field_name_list = copy.copy(self._field_name_list)
        for fname in field_name_list:
            if(fname not in keep_fields):
                self.remove_field(fname)

    def sort_by_field(self, name):
        """Sorts the data along the given field name in ascending order.

        Parameters
        ----------
        name : str
            The name of the field along the events should get sorted.

        Returns
        -------
        sorted_idxs : (n_events,)-shaped numpy ndarray
            The numpy ndarray holding the indices of the sorted array.

        Raises
        ------
        KeyError
            If the given data field does not exist.
        """
        if(name not in self._data_fields):
            raise KeyError('The data field "{}" does not exist in this '
                'DataFieldRecordArray instance!'.format(name))

        sorted_idxs = np.argsort(self._data_fields[name])

        for fname in self._field_name_list:
            self._data_fields[fname] = self._data_fields[fname][sorted_idxs]

        return sorted_idxs

register_FileLoader(['.npy'], NPYFileLoader)
register_FileLoader(['.pkl'], PKLFileLoader)
